"""Verification of the correctness of the step-by-step execution of chunked
prefill with prefix caching. It does so by comparing, at every engine step
(i.e. prefill or decode iteration), a bunch of attributes.
This allows a finer testing of the padding and scheduling implementation.

Run `python -m pytest tests/e2e/test_spyre_pc_inference_steps.py`.
"""

import pytest
from scheduling_utils import (
    create_request_for_scheduler_test,
    random_prompt,
    validate_scheduler_steps,
)
from spyre_util import ModelInfo, verify_block_tables, verify_slot_mappings


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_prefix_hit_within_batch(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two equal sequences are scheduled.
    While prefilling the second sequence we have a prefix cache
    hit and can reuse the first chunk. Note that the fetched prefix blocks
    are still part of the existing decode batch. Hence we have duplicated
    blocks in the block table for this example.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 192,  max tokens = 2, step joining = 0
            * 1: len = 192, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt = random_prompt(model=model, seed=0, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 192,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3]},
            "block_ref_count": {1: 1, 2: 1, 3: 1},
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 192,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3]},
            "block_ref_count": {1: 1, 2: 1, 3: 1},
        },
        {  # prefill chunk 1 seq 1
            # prefix hit!
            "step": 3,
            "tkv": 192,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 1,
            # each chunk has two blocks. Due to padding, the first chunk has
            # only one usable block
            "n_cached_blocks": 1,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3]},
            "block_ref_count": {1: 2, 2: 2, 3: 2},
        },
        {  # prefill chunk 2 seq 1
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 4,
            "tkv": 192,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "n_cached_blocks": 1,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3]},
            "block_ref_count": {1: 2, 2: 2, 3: 2},
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 5,
            "tkv": 193,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 5,
            "n_cached_blocks": 1,
            "block_tables": {"0": [1, 2, 3, 4], "1": [1, 2, 3, 5]},
            "block_ref_count": {1: 2, 2: 2, 3: 2, 4: 1, 5: 1},
        },
        {
            # Tkv should be cleared one step later
            "step": 6,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
            "block_tables": {},
            "block_ref_count": {},
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_block_deduplication_within_batch(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    set_random_seed,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two equal sequences are scheduled. As both sequences
    fit in a single chunk they have to be recomputed. However, we can write
    the KV cache into the same first block as the prompts are identical.
    Therefore we end up with a duplicated block in the block table despite
    not having a prefix hit for this example.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 70, max tokens = 2, step joining = 0
            * 1: len = 70, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt = random_prompt(model=model, seed=0, length=70)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 70,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 2,
            "n_prefix_hits": 0,
            "n_cached_blocks": 0,
            "block_tables": {"0": [1, 2]},
            "block_ref_count": {1: 1, 2: 1},
            "prefill_slot_mappings": {"0": [1, 2]},
        },
        {  # prefill chunk 1 seq 1
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 2,
            "tkv": 70,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "n_cached_blocks": 0,
            "block_tables": {"0": [1, 2], "1": [1, 3]},
            "block_ref_count": {1: 2, 2: 1, 3: 1},
            "prefill_slot_mappings": {
                "1": [0, 3]  # Block 1 is masked out during prefill so it is read-only
            },
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 3,
            "tkv": 71,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 3,
            "block_tables": {"0": [1, 2], "1": [1, 3]},
            "block_ref_count": {1: 2, 2: 1, 3: 1},
        },
        {
            # Tkv should be cleared one step later
            "step": 4,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables, verify_slot_mappings],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_prefix_hit_decoded_block_within_batch(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two sequences are scheduled. We set the second
    sequence to be the entire first sequence plus some generated tokens.
    While prefilling the second sequence we have a prefix cache
    hit and can reuse the first chunk which consists of two blocks. The first
    block is entirely prompt while the second block is a mix of prompt and
    decoded tokens. Note that the fetched prefix blocks are still part of the
    existing decode batch. Hence we have duplicated blocks in the block table
    for this example.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 126,  max tokens = 68, step joining = 0
            * 1: len = 193, max tokens = 2, step joining = 67
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt = random_prompt(model=model, seed=0, length=126)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=68,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    # Next prompt uses part of the first request's output, matching 128 tokens
    # (2 blocks) in total.
    # prompt_len = 126 + 2 + 65 = 193
    prompt2 = (
        prompt
        + list(request1.hf_output["token_ids"][:2])
        + random_prompt(model=model, seed=0, length=65)
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=67,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 126,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 2,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 0.
            "step": 2,
            "tkv": 127,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 2,
        },
        {
            # Decode 3 of request 0.
            # need an additional block
            "step": 4,
            "tkv": 129,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
        },
        {
            # Decode 66 of request 0.
            "step": 67,
            "tkv": 192,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
        },
        {  # prefill chunk 1 seq 1
            # prefix hit of the two blocks in the first chunk
            "step": 68,
            # use unchanged tkv of seq 0, since seq 1 is still prefilling
            "tkv": 192,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 5,
            "n_prefix_hits": 1,
            # 1st block (prompt)
            # 2nd block (prompt + 2 decodes) <- what we want to test
            "n_cached_blocks": 2,
        },
        {  # prefill chunk 2 seq 1
            # no prefix hit, always recompute last chunk
            "step": 69,
            # seq 1 tkv (193) is in 4th block. Need to pad seq 0 tkv to 4th
            # block as well: 192 + 64 = 256
            "tkv": 256,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 5,
            "n_prefix_hits": 0,
            "n_cached_blocks": 2,
            "block_tables": {
                "0": [1, 2, 3],
                "1": [1, 2, 4, 5],
                # Note: new block id 4 instead of 3 here as vLLM does not
                # currently deduplicate decoded blocks and so do we:
                # https://github.com/vllm-project/vllm/blob/1166c31cc78073378a16509fbbbed4cb4f040a4d/vllm/v1/core/block_pool.py#L46
            },
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 70,
            "tkv": 194,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 6,
            "n_cached_blocks": 2,
            "block_tables": {
                "0": [1, 2, 3, 6],
                "1": [1, 2, 4, 5],
            },
        },
        {
            # Tkv should be cleared one step later
            "step": 71,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_prefix_hit_not_in_batch(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two equal sequences are scheduled.
    While prefilling the second sequence we have a prefix cache
    hit and can reuse the first chunk. Note that the fetched prefix blocks
    are not part of the existing decode batch as the sequence has already
    left the batch at the time of prefilling the new sequence. Hence we have
    no duplicated blocks in the block table for this example.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 192,  max tokens = 2, step joining = 0
            * 1: len = 192, max tokens = 2, step joining = 3
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt = random_prompt(model=model, seed=0, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=3,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {
                "0": [1, 2, 3],
            },
        },
        {
            # Decode 1 of request 0.
            # request 1 joined the waiting queue
            "step": 3,
            "tkv": 193,
            "waiting": ["1"],
            "running": [],
            "request_outputs": ["0"],
            "finished_requests": ["0"],
            "n_used_blocks": 4,
        },
        {  # prefill chunk 1 seq 1
            # prefix hit!
            "step": 4,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 1,
            "n_cached_blocks": 1,
        },
        {  # prefill chunk 2 seq 1
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 5,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "n_cached_blocks": 1,
            "block_tables": {
                "1": [1, 2, 3],
            },
        },
        {
            # Decode 1 of request 0.
            "step": 6,
            "tkv": 193,
            "waiting": [],
            "running": [],
            "request_outputs": ["1"],
            "finished_requests": ["1"],
            "n_used_blocks": 4,
            "n_cached_blocks": 1,
        },
        {
            # Tkv should be cleared one step later
            "step": 7,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [4])
def test_limit_blocks_no_prefix_hit(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where three sequences are scheduled with the 1st and 3rd
    sequences being identical. While prefilling the third sequence we don't
    have a prefix cache hit for the first chunk as the KV cache has already
    been overwritten. This is because we limit the number of available blocks
    to 4. Note: When increasing the number of available blocks to 8, see
    test_limit_blocks_prefix_hit, the same test results in a prefix hit.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 3
            * 0: len = 192,  max tokens = 2, step joining = 0
            * 1: len = 192, max tokens = 2, step joining = 3
            * 2: len = 192, max tokens = 2, step joining = 6
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt1 = random_prompt(model=model, seed=0, length=192)
    prompt2 = random_prompt(model=model, seed=1, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=3,
        max_tokens=2,
        prompt=prompt2,  # 1st and 3rd sequence are the same
        use_golden_token_injection=True,
    )

    request3 = create_request_for_scheduler_test(
        model=model,
        request_id=2,
        add_step=6,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 0
            # request 1 joined the waiting queue
            "step": 3,
            "tkv": 193,
            "waiting": ["1"],
            "running": [],
            "request_outputs": ["0"],
            "finished_requests": ["0"],
            "n_used_blocks": 4,
        },
        {  # prefill chunk 1 seq 1
            "step": 4,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 1
            "step": 5,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 1
            # request 2 joined the waiting queue
            "step": 6,
            "tkv": 193,
            "waiting": ["2"],
            "running": [],
            "request_outputs": ["1"],
            "finished_requests": ["1"],
            "n_used_blocks": 4,
        },
        {  # prefill chunk 1 seq 2
            # no prefix hit as KV cache is already overwritten!
            "step": 7,
            "tkv": 192,
            "waiting": [],
            "running": ["2"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 2
            "step": 8,
            "tkv": 192,
            "waiting": [],
            "running": ["2"],
            "request_outputs": ["2"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 2
            "step": 9,
            "tkv": 193,
            "waiting": [],
            "running": [],
            "request_outputs": ["2"],
            "finished_requests": ["2"],
            "n_used_blocks": 4,
        },
        {
            # Tkv should be cleared one step later
            "step": 10,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2, request3],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [4])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_double_prefix_hit_within_batch(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where three equal and one different sequences are scheduled.
    While prefilling the second and fourth sequence we have a prefix cache
    hit and can reuse the first chunk. Note that the fetched prefix blocks
    are still part of the existing decode batch. Hence we have duplicated
    blocks in the block table for this example. More specifically, three
    sequences in the decode batch share the same KV cache block.

    Configuration:
        * max_num_seqs: 4
        * number of prompts: 4
            * 0: len = 192,  max tokens = 2, step joining = 0
            * 1: len = 192, max tokens = 2, step joining = 0
            * 2: len = 192, max tokens = 2, step joining = 0
            * 3: len = 192, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt1 = random_prompt(model=model, seed=0, length=192)
    prompt2 = random_prompt(model=model, seed=1, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request3 = create_request_for_scheduler_test(
        model=model,
        request_id=2,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,  # This request has a different prompt
        use_golden_token_injection=True,
    )

    request4 = create_request_for_scheduler_test(
        model=model,
        request_id=3,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1", "2", "3"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 192,
            "waiting": ["1", "2", "3"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3]},
            "block_ref_count": {1: 1, 2: 1, 3: 1},
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 192,
            "waiting": ["1", "2", "3"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3]},
            "block_ref_count": {1: 1, 2: 1, 3: 1},
        },
        {  # prefill chunk 1 seq 1
            # prefix hit!
            "step": 3,
            "tkv": 192,
            "waiting": ["2", "3"],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 1,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3]},
            "block_ref_count": {1: 2, 2: 2, 3: 2},
        },
        {  # prefill chunk 2 seq 1
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 4,
            "tkv": 192,
            "waiting": ["2", "3"],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3]},
            "block_ref_count": {1: 2, 2: 2, 3: 2},
            "prefill_slot_mappings": {"1": [0, 0]},  # Fully masked prefill
        },
        {  # prefill chunk 1 seq 2
            "step": 5,
            "tkv": 192,
            "waiting": ["3"],
            "running": ["2", "1", "0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3], "2": [4, 5, 6]},
            "block_ref_count": {1: 2, 2: 2, 3: 2, 4: 1, 5: 1, 6: 1},
        },
        {  # prefill chunk 2 seq 2
            "step": 6,
            "tkv": 192,
            "waiting": ["3"],
            "running": ["2", "1", "0"],
            "request_outputs": ["2"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3], "2": [4, 5, 6]},
            "block_ref_count": {1: 2, 2: 2, 3: 2, 4: 1, 5: 1, 6: 1},
        },
        {  # prefill chunk 1 seq 3
            # prefix hit!
            "step": 7,
            "tkv": 192,
            "waiting": [],
            "running": ["3", "2", "1", "0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 1,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3], "2": [4, 5, 6], "3": [1, 2, 3]},
            "block_ref_count": {1: 3, 2: 3, 3: 3, 4: 1, 5: 1, 6: 1},
        },
        {  # prefill chunk 2 seq 3
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 8,
            "tkv": 192,
            "waiting": [],
            "running": ["3", "2", "1", "0"],
            "request_outputs": ["3"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 2, 3], "1": [1, 2, 3], "2": [4, 5, 6], "3": [1, 2, 3]},
            "block_ref_count": {1: 3, 2: 3, 3: 3, 4: 1, 5: 1, 6: 1},
        },
        {
            # Decode 1 of request 0, 1, 2, 3
            "step": 9,
            "tkv": 193,
            "waiting": [],
            "running": [],
            "request_outputs": ["3", "2", "1", "0"],
            "finished_requests": ["3", "2", "1", "0"],
            "n_used_blocks": 10,
            "block_tables": {
                "0": [1, 2, 3, 7],
                "1": [1, 2, 3, 8],
                "2": [4, 5, 6, 9],
                "3": [1, 2, 3, 10],
            },
            "block_ref_count": {1: 3, 2: 3, 3: 3, 7: 1, 8: 1, 4: 1, 5: 1, 6: 1, 9: 1, 10: 1},
        },
        {
            # Tkv should be cleared one step later
            "step": 10,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2, request3, request4],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables, verify_slot_mappings],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [256])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [8])
def test_limit_blocks_prefix_hit(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where three sequences are scheduled with the 1st and 3rd
    sequences being identical. While prefilling the third sequence we
    have a prefix cache hit for the first chunk as the KV cache is still
    persistent. This is because the number of available blocks (8) is high
    enough. Note: When decreasing the number of available blocks to 4, see
    test_limit_blocks_no_prefix_hit, the same test results in a no prefix hit.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 3
            * 0: len = 192,  max tokens = 2, step joining = 0
            * 1: len = 192, max tokens = 2, step joining = 3
            * 2: len = 192, max tokens = 2, step joining = 6
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt1 = random_prompt(model=model, seed=0, length=192)
    prompt2 = random_prompt(model=model, seed=1, length=192)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=3,
        max_tokens=2,
        prompt=prompt2,  # 1st and 3rd sequence are the same
        use_golden_token_injection=True,
    )

    request3 = create_request_for_scheduler_test(
        model=model,
        request_id=2,
        add_step=6,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 192,
            "waiting": [],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 0
            # request 1 joined the waiting queue
            "step": 3,
            "tkv": 193,
            "waiting": ["1"],
            "running": [],
            "request_outputs": ["0"],
            "finished_requests": ["0"],
            "n_used_blocks": 4,
        },
        {  # prefill chunk 1 seq 1
            "step": 4,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 1
            "step": 5,
            "tkv": 192,
            "waiting": [],
            "running": ["1"],
            "request_outputs": ["1"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
        },
        {
            # Decode 1 of request 1
            # request 2 joined the waiting queue
            "step": 6,
            "tkv": 193,
            "waiting": ["2"],
            "running": [],
            "request_outputs": ["1"],
            "finished_requests": ["1"],
            "n_used_blocks": 4,
        },
        {  # prefill chunk 1 seq 2
            # prefix hit as KV cache is still persistent
            "step": 7,
            "tkv": 192,
            "waiting": [],
            "running": ["2"],
            "request_outputs": [],
            "n_used_blocks": 3,
            "n_prefix_hits": 1,
            "n_cached_blocks": 1,
        },
        {  # prefill chunk 2 seq 2
            "step": 8,
            "tkv": 192,
            "waiting": [],
            "running": ["2"],
            "request_outputs": ["2"],
            "n_used_blocks": 3,
            "n_prefix_hits": 0,
            "n_cached_blocks": 1,
        },
        {
            # Decode 1 of request 2
            "step": 9,
            "tkv": 193,
            "waiting": [],
            "running": [],
            "request_outputs": ["2"],
            "finished_requests": ["2"],
            "n_used_blocks": 4,
            "n_cached_blocks": 1,
        },
        {
            # Tkv should be cleared one step later
            "step": 10,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2, request3],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [512])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_multi_chunk_full_match(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two equal sequences are scheduled.
    Both sequences have exactly 3 chunks worth of tokens, thus
    resulting in a 100% match up to the last token. This test
    makes sure that the last chunk is not reused.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 384, max tokens = 2, step joining = 0
            * 1: len = 384, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    prompt = random_prompt(model=model, seed=0, length=384)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 3 seq 0
            "step": 3,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
            # up until this point nothing interesting happened
            # with the block table
            "block_tables": {"0": [1, 2, 3, 4, 5, 6]},
            "block_ref_count": {1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1},
        },
        {  # prefill chunks 1+2 seq 1
            # prefix hit!
            "step": 4,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 1,
            # The number of cached blocks is determined up front
            "n_cached_blocks": 4,  # can reuse the first two chunk (4 blocks)
            # Now, although the last chunk has to be recomputed,
            # the blocks are still shared.
            "block_tables": {"0": [1, 2, 3, 4, 5, 6], "1": [1, 2, 3, 4, 5, 6]},
            "block_ref_count": {1: 2, 2: 2, 3: 2, 4: 2, 5: 2, 6: 2},
        },
        {  # prefill chunk 3 seq 1
            # cannot use prefix, as the last chunk has to always be recomputed
            "step": 5,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
            "n_cached_blocks": 4,
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 6,
            "tkv": 385,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 8,
            "n_cached_blocks": 4,
            # when decode starts, we see the tables diverge
            "block_tables": {"0": [1, 2, 3, 4, 5, 6, 7], "1": [1, 2, 3, 4, 5, 6, 8]},
            "block_ref_count": {1: 2, 2: 2, 3: 2, 4: 2, 5: 2, 6: 2, 7: 1, 8: 1},
        },
        {
            # Tkv should be cleared one step later
            "step": 7,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [512])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_multi_chunk_partial_match_misaligned(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two sequences are scheduled which share a common
    prefix. The second sequence shares 254 tokens with the first sequence,
    which is less than two chunks. We can therefore reuse only one chunk
    (254 < 2*128 = 256). This leads to computation of the entire second chunk,
    including the recomputation of the third block even though we already
    have it in cache.

    p1 = [AB|CD|EF]
    p2 = [AB|CX|EF]

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 384,  max tokens = 2, step joining = 0
            * 1: len = 384, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    # twice the same seed for a sequence of length 384
    # the first sequence shares the same prefix of length 384 tokens
    # the second sequence shares the same prefix of length 254 tokens
    # hence sequence 1 shares the first 254 tokens with sequence 0

    prompt1 = random_prompt(model=model, seed=0, length=384)
    prompt2 = prompt1[0:254] + random_prompt(model=model, seed=0, length=384 - 254)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 3 seq 0
            "step": 3,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 1 seq 1
            # prefix hit!
            "step": 4,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
            "n_prefix_hits": 1,
            # The number of cached blocks is determined up front
            "n_cached_blocks": 2,  # can only reuse the first chunk (2 blocks)
            "prefill_slot_mappings": {"1": [1, 2]},
        },
        {  # prefill chunk 2 seq 1
            # cannot use prefix, as the prefix is less than 2 chunks
            "step": 5,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 9,
            "n_prefix_hits": 0,
            "n_cached_blocks": 2,
            "prefill_slot_mappings": {"1": [0, 4]},  # Block 3 (prefix hit) is masked out
        },
        {  # prefill chunk 3 seq 1
            "step": 6,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 9,
            "n_prefix_hits": 0,
            "n_cached_blocks": 2,
            "block_tables": {
                "0": [1, 2, 3, 4, 5, 6],
                "1": [1, 2, 3, 7, 8, 9],
            },
            "prefill_slot_mappings": {"1": [8, 9]},
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 7,
            "tkv": 385,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 11,
            "n_cached_blocks": 2,
        },
        {
            # Tkv should be cleared one step later
            "step": 8,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables, verify_slot_mappings],
    )


@pytest.mark.chunked_prefill
@pytest.mark.full_model
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [512])
@pytest.mark.parametrize("max_num_batched_tokens", [128])
@pytest.mark.parametrize("available_blocks", [None])
def test_multi_chunk_partial_match_aligned(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """Scenario where two sequences are scheduled which share a common
    prefix. The second sequence shares 256 tokens with the first sequence,
    which is exactly two chunks. We can therefore reuse both chunks as the
    second chunk is not the last chunk (3rd) which needs to be recomputed.

    Configuration:
        * max_num_seqs: 2
        * number of prompts: 2
            * 0: len = 384, max tokens = 2, step joining = 0
            * 1: len = 384, max tokens = 2, step joining = 0
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    # two sequences spanning exactly three chunks each. The
    # second sequence shares a two chunk prefix with the first

    prompt1 = random_prompt(model=model, seed=0, length=384)
    prompt2 = prompt1[0:256] + random_prompt(model=model, seed=0, length=384 - 256)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill chunk 1 seq 0
            "step": 1,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 2 seq 0
            "step": 2,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": [],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 3 seq 0
            "step": 3,
            "tkv": 384,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 6,
            "n_prefix_hits": 0,
        },
        {  # prefill chunk 1 and 2 seq 1
            # prefix hit!
            "step": 4,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 8,
            "n_prefix_hits": 1,
            # The number of cached blocks is determined up front
            "n_cached_blocks": 4,  # Reusing two chunks (4 blocks)
        },
        {  # prefill chunk 3 seq 1
            "step": 5,
            "tkv": 384,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 8,
            "n_prefix_hits": 0,
            "n_cached_blocks": 4,
            "block_tables": {
                "0": [1, 2, 3, 4, 5, 6],
                "1": [1, 2, 3, 4, 7, 8],
            },
        },
        {
            # Decode 1 of request 0.
            # Decode 1 of request 1.
            "step": 6,
            "tkv": 385,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 10,
            "n_cached_blocks": 4,
        },
        {
            # Tkv should be cleared one step later
            "step": 7,
            "tkv": 0,
            "waiting": [],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables],
    )


@pytest.mark.chunked_prefill
@pytest.mark.prefix_caching
# These values are all parameterized for test sorting
@pytest.mark.parametrize("max_num_seqs", [2])
@pytest.mark.parametrize("max_model_len", [576])
@pytest.mark.parametrize("max_num_batched_tokens", [192])
@pytest.mark.parametrize("available_blocks", [None])
def test_first_chunk_partial_match(
    model: ModelInfo,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    max_num_seqs: int,
    max_model_len: int,
    max_num_batched_tokens: int,
    available_blocks: int,
):
    """There was a bug where a partial match in the first chunk could cause a
    crash. This test covers that case.
    """
    monkeypatch.setenv("VLLM_SPYRE_CP_INTERLEAVE_STEPS", "0")

    # The bug occurred because the sum of the left padding plus the usable prefix cache was not
    # divisible by the chunk size while the prefix cache length was at least one block.
    # To get this to occur, the usable blocks had to be zero and the left padding needed to be
    # non-zero. This can't happen with a chunk size of 2 blocks, because even a single left-pad will
    # cause the first block of prefix cache hit to be "usable". So for this test we create 3-block
    # chunks so that we can have a single left-padding block and a single prefix block hit.

    # Calculate length of two-chunk prompt with a single left pad block
    single_left_pad_prompt_len = (192 * 2) - 64

    # First prompt just one block
    prompt1 = random_prompt(model=model, seed=0, length=64)
    # Second prompt hits the one block of prefix cache with one left pad block
    prompt2 = prompt1 + random_prompt(model=model, seed=0, length=single_left_pad_prompt_len - 64)

    request1 = create_request_for_scheduler_test(
        model=model,
        request_id=0,
        add_step=0,
        max_tokens=2,
        prompt=prompt1,
        use_golden_token_injection=True,
    )

    request2 = create_request_for_scheduler_test(
        model=model,
        request_id=1,
        add_step=0,
        max_tokens=2,
        prompt=prompt2,
        use_golden_token_injection=True,
    )

    checked_steps = [
        {
            "step": 0,
            "tkv": 0,
            "waiting": ["0", "1"],
            "running": [],
            "request_outputs": [],
            "n_used_blocks": 0,
        },
        {  # prefill seq 0
            "step": 1,
            "tkv": 64,
            "waiting": ["1"],
            "running": ["0"],
            "request_outputs": ["0"],
            "n_used_blocks": 1,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1]},
        },
        {  # prefill seq 1. This step was crashing before
            "step": 2,
            "tkv": 64,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": [],
            "n_used_blocks": 5,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1], "1": [1, 2, 3, 4, 5]},
            "prefill_slot_mappings": {
                "1": [0, 0, 2]  # One mask for left padding, one mask for block `1` to not be
                # overwritten since it hit cache
            },
        },
        {  # finish seq 1 prefill
            "step": 3,
            "tkv": 320,
            "waiting": [],
            "running": ["1", "0"],
            "request_outputs": ["1"],
            "n_used_blocks": 5,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1], "1": [1, 2, 3, 4, 5]},
            "prefill_slot_mappings": {"1": [3, 4, 5]},
        },
        {  # decode
            "step": 4,
            "tkv": 321,
            "waiting": [],
            "running": [],
            "request_outputs": ["1", "0"],
            "finished_requests": ["1", "0"],
            "n_used_blocks": 7,
            "n_prefix_hits": 0,
            "block_tables": {"0": [1, 6], "1": [1, 2, 3, 4, 5, 7]},
        },
    ]

    validate_scheduler_steps(
        model=model,
        backend=backend,
        monkeypatch=monkeypatch,
        requests=[request1, request2],
        checked_steps=checked_steps,
        max_num_seqs=max_num_seqs,
        max_model_len=max_model_len,
        available_blocks=available_blocks,
        max_num_batched_tokens=max_num_batched_tokens,
        prefix_caching=True,
        extra_assert_funcs=[verify_block_tables, verify_slot_mappings],
    )



import pytest
import requests
from sentence_transformers import CrossEncoder
from spyre_util import get_spyre_backend_list, get_spyre_model_list


@pytest.mark.parametrize("model", get_spyre_model_list(isScoring=True))
@pytest.mark.parametrize(
    "warmup_shapes",
    [  # (prompt_length/new_tokens/batch_size)
        pytest.param([(64, 0, 4)]),
    ],
)
@pytest.mark.parametrize("backend", get_spyre_backend_list())
@pytest.mark.scoring
def test_serving(remote_openai_server, model, warmup_shapes, backend):
    """Test online serving using the `vllm serve` CLI"""

    score_url = remote_openai_server.url_for("/score")

    query = "What is the capital of France?"
    # Number of inputs larger than the warmup batch size of 4
    # and with a non-uniform token length
    docs = [
        "The capital of France is Paris.",
        "The capital of Germany is Berlin.",
        "The capital of Brazil is Brasilia.",
        "The capital of the country with the best beer is Berlin.",
        "The capital of the USA is Washington.",
        "The capital city of Spain is Madrid.",
    ]
    vllm_outputs = requests.post(
        url=score_url,
        json={
            "text_1": query,
            "text_2": docs,
        },
    ).json()
    vllm_scores = [o["score"] for o in vllm_outputs["data"]]

    ce_model = CrossEncoder(model.name, revision=model.revision)
    ce_scores = ce_model.predict(
        [
            (query, docs[0]),
            (query, docs[1]),
            (query, docs[2]),
            (query, docs[3]),
            (query, docs[4]),
            (query, docs[5]),
        ]
    )

    assert ce_scores[0] == pytest.approx(vllm_scores[0], rel=0.02)
    assert ce_scores[1] == pytest.approx(vllm_scores[1], rel=0.02)
    assert ce_scores[2] == pytest.approx(vllm_scores[2], rel=0.02)
    assert ce_scores[3] == pytest.approx(vllm_scores[3], rel=0.02)
    assert ce_scores[4] == pytest.approx(vllm_scores[4], rel=0.02)
    assert ce_scores[5] == pytest.approx(vllm_scores[5], rel=0.02)



from vllm_spyre.multimodal.mm_mappings.base import MMUtilsBase, MMWarmupInputs
from vllm_spyre.multimodal.mm_mappings.llava_next import LlavaNextMMUtils

__all__ = ["MMWarmupInputs", "MMUtilsBase", "LlavaNextMMUtils"]



"""Utilities for selecting and loading Spyre models."""

import os
from dataclasses import dataclass
from typing import Any, cast

import torch
import torch._inductor.config
import torch.distributed as dist
import torch.nn as nn
from fms.models import get_model
from transformers import PretrainedConfig
from vllm.config import ModelConfig, VllmConfig
from vllm.forward_context import get_forward_context
from vllm.logger import init_logger
from vllm.model_executor.model_loader.weight_utils import download_weights_from_hf
from vllm.v1.outputs import SamplerOutput
from vllm.v1.sample.metadata import SamplingMetadata
from vllm.v1.sample.sampler import Sampler

import vllm_spyre.envs as envs_spyre
import vllm_spyre.multimodal as spyre_mm
import vllm_spyre.utils as utils_spyre
from vllm_spyre.platform import SpyrePlatform

try:
    import backends.dynamo_tracer  # ty: ignore[unresolved-import] # noqa
except ImportError:
    print("WARNING: Disabled: dynamo_tracer")
    pass

BACKEND_LIST = ["sendnn", "inductor"]

logger = init_logger(__name__)


@dataclass
class SpyreAttentionMetadata:
    slot_mapping: torch.Tensor
    current_tkv_mask: torch.Tensor
    left_padded_prompt_mask: torch.Tensor
    block_table: torch.Tensor
    is_prefill: bool
    # We need this indices because when requests are removed from the
    # persistent batch, we need to keep the reference of the remaining
    # requests, that is, this index must be the same from the prefill until
    # the end.
    scale_indices: torch.Tensor


class SpyreCausalLM(nn.Module):
    def __init__(
        self,
        vllm_config: VllmConfig,
        max_prompt_length: int,
        max_decode_length: int,
        rank: int,
    ) -> None:
        super().__init__()

        self.sampler = Sampler()

        # boolean tensor of length batch size with indices:
        # True for unfinished sequences and
        # False for finished or padded sequences
        self.indices = None

        # number of right pads (relevant for continuous batching only)
        self.n_pads_right = 0

        self._mask_dtype = (
            torch.float16 if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND == "sendnn" else torch.float32
        )

        # FMS Model
        if envs_spyre.VLLM_SPYRE_USE_CB:
            self.model = ContinuousBatchingFmsModel(vllm_config, rank)
        else:
            self.model = StaticBatchingFmsModel(
                vllm_config,
                max_prompt_length,
                max_decode_length,
                rank,
            )

        # Pull mm model utils / is_multimodal up a level for convenience
        self.mm_model_utils = self.model.mm_model_utils
        self.is_multimodal = self.model.is_multimodal

    def forward(
        self,
        input_ids_or_embeds: torch.Tensor,
        positions: torch.Tensor,
        masks: torch.Tensor,
        is_prompt: bool,
    ) -> torch.Tensor:
        if is_prompt and not envs_spyre.VLLM_SPYRE_USE_CB:
            self.model.past_key_value_states = None

        extra_kwargs: dict[str, Any] = {}
        if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND != "sendnn":
            # Bug in 2.3.1 fixed in 2.4.1 for SDPA flash
            # cpu impl when padding too much
            extra_kwargs["attn_algorithm"] = "math"

        # normal prefill or decoding step
        logits = self.model(
            input_ids_or_embeds,
            position_ids=positions,
            mask=masks,
            use_cache=True,
            is_prompt=is_prompt,
            **extra_kwargs,
        )

        if envs_spyre.VLLM_SPYRE_USE_CB:
            if is_prompt and self.n_pads_right > 0:
                # get last token before the right padding
                logits = logits[self.indices, -self.n_pads_right - 1, :]
            else:
                # just take last token if no right padding
                logits = logits[self.indices, -1, :]
        else:
            # removing finished or padded sequences
            logits = logits[self.indices]

        return logits

    def get_maybe_mm_embeddings(self, input_ids, mm_features, is_decode):
        """If the model is multimodal, get the (maybe) multimodal embeddings.
        If it isn't, return None, since we only use embeddings for multimodal.

        In the case of prefill / decode; we should only have mm features in
        prefill, because by that point, the multimodal data will already be
        merged into the embeddings to be cached. As such we explicitly explode
        if mm_features are passed in decode, because it's likely a mistake.

        NOTE: generally is_decode will set iteration > 0 in FMS; failing to do
        this will cause it to return the raw input id and try to embed in the
        forward call, which may break on AIU due to misalignment with prefill's
        embedding call.
        """
        if not self.model.is_multimodal or self.mm_model_utils is None:
            # The model is likely implemented incorrectly or not initialized,
            # or we are passing multimodal features to a model that should not
            # take them.
            if mm_features:
                raise ValueError("mm_features were provided, but model is not multimodal!")
            # We do not use embeddings for models that aren't multimodal.
            return None

        # Delegate to this model architecture's multimodal helpers to
        # get the (potentially) multimodal embeddings from the FMS model.
        fms_model = self.model.model
        return self.mm_model_utils.get_maybe_mm_embeddings(
            fms_model,
            input_ids,
            mm_features,
            is_decode,
        )

    def sample(
        self,
        logits: torch.Tensor,
        sampling_metadata: SamplingMetadata,
    ) -> SamplerOutput | None:
        next_tokens = self.sampler(logits, sampling_metadata)
        return next_tokens

    def get_mask_dtype(self) -> torch.dtype:
        return self._mask_dtype


class FmsModelBase(nn.Module):
    def __init__(
        self,
        vllm_config: VllmConfig,
        max_prompt_length: int,
        max_decode_length: int,
        rank: int,
        sendnn_dynamic: bool,
    ) -> None:
        super().__init__()

        self.config: PretrainedConfig = vllm_config.model_config.hf_config

        # Actual FMS model
        self.model: nn.Module
        self.model_config = vllm_config.model_config
        self.parallel_config = vllm_config.parallel_config
        self.cache_config = vllm_config.cache_config
        self.scheduler_config = vllm_config.scheduler_config
        self.dtype = self.get_dtype()

        # Wrappers for utils for multimodal
        self.mm_model_utils: spyre_mm.MMUtilsBase | None = None
        self.is_multimodal = False

        # Load the weights from the cached or downloaded files.
        self.load_weights(
            model_config=self.model_config,
            max_prompt_length=max_prompt_length,
            max_decode_length=max_decode_length,
            distributed_strategy="tp" if self.parallel_config.world_size > 1 else None,
            sendnn_dynamic=sendnn_dynamic,
            rank=rank,
            world_size=self.parallel_config.world_size,
        )

    def get_dtype(self) -> torch.dtype:
        raise NotImplementedError()

    def load_weights(
        self,
        model_config: ModelConfig,
        max_prompt_length: int,
        max_decode_length: int,
        distributed_strategy: str | None,
        sendnn_dynamic: bool,
        **kwargs,
    ) -> None:
        logger.debug("Loading model weights for model %s", model_config.model)
        logger.debug("Model config has dtype: %s", model_config.dtype)

        # When using quantized models, we might not be using the
        # model_config's dtype, hence we don't log the msg below
        # since it might confuse the user
        if model_config.quantization:
            logger.debug("Quantized model found with quantization : %s", model_config.quantization)
        else:
            if self.dtype is not model_config.dtype:
                logger.info(
                    "Ignoring user-provided dtype=%s (provided either through"
                    " --dtype CLI arg or model_config.dtype) and using"
                    " dtype=%s instead.",
                    model_config.dtype,
                    self.dtype,
                )

        is_local = os.path.isdir(model_config.model)
        model_path = model_config.model
        # Get location of model from HF cache.
        if not is_local:
            model_path = download_weights_from_hf(
                model_name_or_path=model_path,
                cache_dir=None,
                allow_patterns=["*.safetensors", "*.bin", "*.pt"],
                revision=model_config.revision,
            )

        # Get any fixes needed that must be patched into the kwargs;
        # currently this is only use for multimodal models / llava next
        model_kwargs = spyre_mm.get_mm_specific_load_overrides(self.config)

        with utils_spyre.stagger_region(
            envs_spyre.VLLM_SPYRE_MAX_LOAD_PROCESSES,
            kwargs["world_size"],
            kwargs["rank"],
        ):
            self.model = get_model(
                architecture="hf_pretrained",
                model_path=model_path,
                distributed_strategy=distributed_strategy,
                group=dist.group.WORLD,
                fused_weights=False,
                **model_kwargs,
            )

        self.model.eval()
        torch.set_grad_enabled(False)

        _target_cache_size = max(int(max_decode_length * 2), int(max_prompt_length * 2.5))
        if (
            hasattr(torch._dynamo.config, "accumulated_cache_size_limit")
            and _target_cache_size > torch._dynamo.config.accumulated_cache_size_limit
        ):
            _prev = torch._dynamo.config.accumulated_cache_size_limit
            torch._dynamo.config.accumulated_cache_size_limit = _target_cache_size
            logger.info(
                "NOTICE: Adjusting "
                "torch._dynamo.config.accumulated_cache_size_limit "
                "from %s to %s "
                "to accommodate prompt size of %d "
                "and decode tokens of %d",
                _prev,
                torch._dynamo.config.accumulated_cache_size_limit,
                max_prompt_length,
                max_decode_length,
            )

        if _target_cache_size > torch._dynamo.config.cache_size_limit:
            _prev = torch._dynamo.config.cache_size_limit
            torch._dynamo.config.cache_size_limit = _target_cache_size
            logger.info(
                "NOTICE: Adjusting torch._dynamo.config.cache_size_limit "
                "from %s to %s "
                "to accommodate prompt size of %d "
                "and decode tokens of %d",
                _prev,
                torch._dynamo.config.accumulated_cache_size_limit,
                max_prompt_length,
                max_decode_length,
            )

        if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND in BACKEND_LIST:
            # When running on Spyre cards for either non-quantized (bf16) models
            # or quantized (fp8) models, we cast any bf16 params down
            self._cast_bf16_to_f16()
            options = {"sendnn.dynamic": True} if sendnn_dynamic else {}

            # Lazy import to avoid load torch_sendnn runtime before it is really
            # necessary. This solve issues of running forked tests that share
            # some resources from parent to children which can have problems
            # of caching even though the test run in isolated subprocesses.

            if SpyrePlatform.sendnn_configured():
                pass

            self.model = torch.compile(
                self.model,
                backend=envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND,
                options=options,
            )
        else:
            # CPU execution
            # For continuous batching w/ paged attention, we only support either
            # fp32 or fp8, not f16 or bf16.
            if not model_config.quantization:
                assert self.dtype == torch.float32
                self._cast_to_f32()

        # If it's multimodal, create an instance of the
        # corresponding mm utils helper; this is arch specific.
        self.mm_model_utils = spyre_mm.maybe_get_mm_utils(
            model_path=model_path,
            fms_config=self.model.config,
            hf_config=self.config,
        )
        self.is_multimodal = self.mm_model_utils is not None
        logger.debug("Model weights loaded successfully.")

    def _cast_bf16_to_f16(self):
        """Cast all bf16 params in the model to f16."""
        for name, param in self.model.named_parameters():
            if param.dtype == torch.bfloat16:
                logger.debug(
                    "You are casting param %s to fp16, which"
                    " will cause loss of accuracy. This is required for"
                    " spyre cards that don't support bf16. You can ignore"
                    " this warning if this is intended.",
                    name,
                )
                param.data = param.data.to(dtype=torch.float16)

    def _cast_to_f32(self):
        """Cast model parameters to f32."""
        for name, param in self.model.named_parameters():
            logger.debug(
                "Casting param %s to fp32. This is required"
                " for attention implementations that only support"
                " full precision.",
                name,
            )
            param.data = param.data.to(dtype=torch.float32)


class ContinuousBatchingFmsModel(FmsModelBase):
    def __init__(
        self,
        vllm_config: VllmConfig,
        rank: int,
    ) -> None:
        BLOCK_SIZE = SpyrePlatform.get_block_size()
        max_model_len = vllm_config.model_config.max_model_len

        # edge case: prompt fills model length: can produce 1 token with prefill
        max_prompt_length = max_model_len
        # edge case: prompt will be padded to first block:
        # can produce 1 token with prefill plus rest of model length
        max_decode_length = max_model_len - BLOCK_SIZE + 1

        super().__init__(
            vllm_config, max_prompt_length, max_decode_length, rank, sendnn_dynamic=True
        )

        self.prefill_past_key_values = None

        # physical KV cache on AIU Spyre: will eventually not live in this class
        self.kv_cache_specs = {}
        self.kv_cache_specs["block_size"] = BLOCK_SIZE
        self.kv_cache_specs["num_kv_heads"] = self.model_config.get_num_kv_heads(
            self.parallel_config
        )

        if self.config.model_type in {"llama", "granite", "granitemoehybrid"}:
            self.kv_cache_specs["num_layers"] = self.config.num_hidden_layers
            self.kv_cache_specs["head_dim"] = getattr(
                self.model.config,
                "head_dim",
                self.config.hidden_size // self.config.num_attention_heads,
            )
        elif self.config.model_type == "gpt_bigcode":
            self.kv_cache_specs["num_layers"] = self.config.n_layer
            self.kv_cache_specs["head_dim"] = self.config.n_embd // self.config.n_head
        elif self.is_multimodal and self.mm_model_utils is not None:
            # Handle multimodal separately for now since we need to unwrap the
            # text configs and technically (outside FMS) the LLM could be
            # generic; the instance of mm_model_utils encapsulates the configs,
            # so no need to pass them again.
            unwrapped_opts = self.mm_model_utils.unwrap_mm_kv_cache_opts()
            self.kv_cache_specs.update(unwrapped_opts)
        else:
            raise NotImplementedError(
                f"[SpyreCausalLM] model type {self.config.model_type} "
                f"not supported in ContinuousBatchingFmsModel"
            )

        if self.model_config.quantization:
            self.attention_name = "spyre_paged_attn_fp8"
            self.is_fp8_model = True
        else:
            self.attention_name = "spyre_paged_attn"
            self.is_fp8_model = False

        self.current_scale: list[tuple] | None = None

    def set_past_key_value_states(self, num_blocks) -> None:
        # List[layers] of Tuple[k,v] of
        # Tensor[num_blocks, block_size, num_kv_heads, head_dim]
        if not self.model_config.quantization:
            self.past_key_value_states = [
                (
                    torch.zeros(
                        num_blocks,
                        self.kv_cache_specs["block_size"],
                        self.kv_cache_specs["num_kv_heads"],
                        self.kv_cache_specs["head_dim"],
                        dtype=self.dtype,
                    ),
                    torch.zeros(
                        num_blocks,
                        self.kv_cache_specs["block_size"],
                        self.kv_cache_specs["num_kv_heads"],
                        self.kv_cache_specs["head_dim"],
                        dtype=self.dtype,
                    ),
                )
                for _ in range(self.kv_cache_specs["num_layers"])
            ]
        else:
            from fms_mo.aiu_addons.fp8.fp8_utils import ScaledTensor

            batch_size = max(2, self.scheduler_config.max_num_seqs)
            self.past_key_value_states = [
                (
                    ScaledTensor(
                        torch.zeros(
                            num_blocks,
                            self.kv_cache_specs["block_size"],
                            self.kv_cache_specs["num_kv_heads"],
                            self.kv_cache_specs["head_dim"],
                            dtype=self.dtype,
                        ),
                        scale=torch.tensor([1.0] * batch_size, dtype=torch.float32),
                        scaled=False,
                    ),
                    ScaledTensor(
                        torch.zeros(
                            num_blocks,
                            self.kv_cache_specs["block_size"],
                            self.kv_cache_specs["num_kv_heads"],
                            self.kv_cache_specs["head_dim"],
                            dtype=self.dtype,
                        ),
                        scale=torch.tensor([1.0] * batch_size, dtype=torch.float32),
                        scaled=False,
                    ),
                )
                for _ in range(self.kv_cache_specs["num_layers"])
            ]
            # This list keep the reference of scales of the quantized weights
            # that will be updated after model execution
            self.current_kv_scales = [
                (k_cache._scale, v_cache._scale) for k_cache, v_cache in self.past_key_value_states
            ]

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor,
        mask: torch.Tensor,
        use_cache: bool,
        is_prompt: bool,
        **extra_kwargs,
    ) -> torch.Tensor:
        forward_context = get_forward_context()

        attn_metadata = cast(SpyreAttentionMetadata, forward_context.attn_metadata)
        assert attn_metadata is not None
        # import will be not be needed/ handled by FMS soon
        import fms.utils.spyre.paged  # noqa # pylint: disable=unused-import

        # specify attention type for continuous batching
        extra_kwargs["attn_name"] = self.attention_name

        if self.is_fp8_model:
            # set scale for kv_cache
            self._set_scale_for_fp8(attn_metadata)

            # Adjust decode for bs=1 if needed
            input_ids, position_ids, attn_metadata = self._adjust_input_for_fp8(
                input_ids=input_ids, position_ids=position_ids, attn_metadata=attn_metadata
            )

        # Run the model
        output = self.model(
            input_ids,
            position_ids=position_ids,
            mask=mask,
            past_key_value_states=self.past_key_value_states,
            use_cache=use_cache,
            last_n_tokens=SpyrePlatform.get_block_size() if is_prompt else 1,
            current_tkv_mask=attn_metadata.current_tkv_mask,
            left_padded_prompt_mask=attn_metadata.left_padded_prompt_mask,
            block_table=attn_metadata.block_table,
            slot_mapping=attn_metadata.slot_mapping,
            **extra_kwargs,
        )

        logits, self.past_key_value_states = output

        if is_prompt:
            # assert that indeed received the last block of logits
            assert logits.shape[1] == SpyrePlatform.get_block_size()

        if self.is_fp8_model:
            # update scale for kv_cache after execute model
            self._update_scale_for_fp8(attn_metadata)

            logits = self._adjust_output_for_fp8(logits, attn_metadata)

        return logits

    def _set_scale_for_fp8(self, attn_metadata: SpyreAttentionMetadata):
        for layer_idx, (k, v) in enumerate(self.past_key_value_states):
            if envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL:
                # static scaling
                if attn_metadata.is_prefill:
                    k._scale = torch.ones(1, dtype=torch.float32)
                    v._scale = torch.ones(1, dtype=torch.float32)
                elif len(attn_metadata.scale_indices) == 1:
                    k._scale = torch.ones(2, dtype=torch.float32)
                    v._scale = torch.ones(2, dtype=torch.float32)
                else:
                    k._scale = torch.ones(len(attn_metadata.scale_indices), dtype=torch.float32)
                    v._scale = torch.ones(len(attn_metadata.scale_indices), dtype=torch.float32)

                # interesting- no mark_dynamic required on decode here

                k._scaled = True
                v._scaled = True
                continue

            if attn_metadata.is_prefill:
                # NOTE: Currently, prefill is only for a single prompt
                # In prefill, we restore the scale (no scale) and
                # reset to 1.
                assert len(attn_metadata.scale_indices) == 1
                prefill_index = attn_metadata.scale_indices[0]
                k._scale = self.current_kv_scales[layer_idx][0][prefill_index] = torch.ones(
                    1, dtype=torch.float32
                )
                v._scale = self.current_kv_scales[layer_idx][1][prefill_index] = torch.ones(
                    1, dtype=torch.float32
                )
                k._scaled = False
                v._scaled = False
            elif len(attn_metadata.scale_indices) == 1:
                # Decode
                # Special case for decode of bs=1, pad the batch to be bs=2
                dec_index = attn_metadata.scale_indices[0]
                k._scale = self.current_kv_scales[layer_idx][0][dec_index].repeat(2)
                v._scale = self.current_kv_scales[layer_idx][1][dec_index].repeat(2)

            else:
                # Set scale only for the requests of the batch
                k._scale = self.current_kv_scales[layer_idx][0][
                    attn_metadata.scale_indices
                ].reshape(-1)
                v._scale = self.current_kv_scales[layer_idx][1][
                    attn_metadata.scale_indices
                ].reshape(-1)

            # We set dynamic only for the first dimension of scale
            # during decoding
            is_dynamic_flag = 0 if attn_metadata.is_prefill else 1

            torch._dynamo.mark_dynamic(v._scale, is_dynamic_flag)
            torch._dynamo.mark_dynamic(k._scale, is_dynamic_flag)

    def _update_scale_for_fp8(self, attn_metadata: SpyreAttentionMetadata):
        if envs_spyre.VLLM_SPYRE_USE_CHUNKED_PREFILL:
            # static scaling
            return

        for layer_idx, (k, v) in enumerate(self.past_key_value_states):
            if attn_metadata.is_prefill or len(attn_metadata.scale_indices) > 1:
                self.current_kv_scales[layer_idx][0][attn_metadata.scale_indices] = k._scale
                self.current_kv_scales[layer_idx][1][attn_metadata.scale_indices] = v._scale
            else:
                # if we did the padding, then we need to update only the scale
                # for the decoding index
                self.current_kv_scales[layer_idx][0][attn_metadata.scale_indices[0]] = k._scale[0]
                self.current_kv_scales[layer_idx][1][attn_metadata.scale_indices[0]] = v._scale[0]

    def get_dtype(self) -> torch.dtype:
        # Get the model's data type
        # This should be:
        # FP32 for un-quantized models on cpu
        # FP16 for un-quantized models on spyre
        # FP8 (float8_e4m3fn) for quantized models
        # (only fp8 quantization is supported)
        if self.model_config.quantization:
            return torch.float8_e4m3fn
        else:
            if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND in BACKEND_LIST:
                return torch.float16
            else:
                return torch.float32

    # TODO: this is not the best place to do. But we expect this to
    # be temporary and here should be easy to remove later
    def _adjust_input_for_fp8(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor,
        attn_metadata: SpyreAttentionMetadata,
    ):
        # NOTE: We only need to adjust the inputs for decode with
        # batch_size=2
        if attn_metadata.is_prefill or input_ids.shape[0] > 1:
            return input_ids, position_ids, attn_metadata

        input_ids = input_ids.repeat(2, 1)
        position_ids = position_ids.repeat(2, 1)
        attn_metadata = SpyreAttentionMetadata(
            slot_mapping=attn_metadata.slot_mapping.repeat(2, 1),
            current_tkv_mask=attn_metadata.current_tkv_mask.repeat(2),
            left_padded_prompt_mask=attn_metadata.left_padded_prompt_mask.repeat(2),
            block_table=attn_metadata.block_table.repeat(2, 1),
            is_prefill=attn_metadata.is_prefill,
            # NOTE: we don't change here, because we'll need this untouched
            # when we update the the scale after run the model
            scale_indices=attn_metadata.scale_indices,
        )
        return input_ids, position_ids, attn_metadata

    def _adjust_output_for_fp8(self, logits: torch.Tensor, attn_metadata: SpyreAttentionMetadata):
        if attn_metadata.is_prefill or len(attn_metadata.scale_indices) > 1:
            # skip for prefill or decode for bs>1
            return logits

        return logits[0].unsqueeze(0)


class StaticBatchingFmsModel(FmsModelBase):
    def __init__(
        self,
        vllm_config: VllmConfig,
        max_prompt_length: int,
        max_decode_length: int,
        rank: int,
    ) -> None:
        super().__init__(
            vllm_config, max_prompt_length, max_decode_length, rank, sendnn_dynamic=False
        )

        # dynamic KV cache
        self.past_key_value_states = None

        if self.model_config.quantization:
            self.attention_name = "math_fp8"
        else:
            self.attention_name = "sdpa_causal"

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor,
        mask: torch.Tensor,
        use_cache: bool,
        **extra_kwargs,
    ) -> torch.Tensor:
        # specify attention type for static batching
        extra_kwargs["attn_name"] = self.attention_name

        # In order to calculate prompt logprobs, we have to return the
        # hidden states from the whole prompt. The static graphs need to be
        # compiled with this set one way or the other.
        last_n_tokens = 0 if envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS else 1

        output = self.model(
            input_ids,
            position_ids=position_ids,
            mask=mask,
            past_key_value_states=self.past_key_value_states,
            use_cache=use_cache,
            last_n_tokens=last_n_tokens,
            **extra_kwargs,
        )

        logits, self.past_key_value_states = output

        if not envs_spyre.VLLM_SPYRE_ENABLE_PROMPT_LOGPROBS:
            logits = logits.squeeze(1)

        return logits

    def get_dtype(self) -> torch.dtype:
        # For static batching, we set fp16 on spyre and fp32 on cpu
        # (This applies even when running fp8 quantized models)
        if envs_spyre.VLLM_SPYRE_DYNAMO_BACKEND in BACKEND_LIST:
            return torch.float16
        else:
            return torch.float32



"""Contains utilities for generating and comparing output and
returning the results"""

import math
import os
from pathlib import Path
from typing import Any, Union

import numpy as np
import pytest
import torch
from hf_result_cache import HFResultCache
from llm_cache import LLM_CACHE, get_cached_llm
from sentence_transformers import SentenceTransformer, util
from spyre_util import DecodeWarmupShapes, EmbeddingWarmupShapes, ModelInfo
from transformers import AutoModelForCausalLM, AutoTokenizer
from vllm import LLM, SamplingParams
from vllm.transformers_utils.tokenizer import get_tokenizer

DISABLE_ASSERTS = False  # used for debugging

ISCLOSE_ABS_TOL = float(os.environ.get("VLLM_SPYRE_TEST_ABS_TOL", "0.08"))
ISCLOSE_ABS_TOL_QUANTIZATION = float(os.environ.get("VLLM_SPYRE_TEST_QUANTIZED_ABS_TOL", "0.17"))

HF_RESULT_CACHE = HFResultCache()


# Hugging Face
def generate_hf_output(
    model: str | ModelInfo,
    prompts: Union[list[str], list[list[int]]],  # also accept token ids
    max_new_tokens: Union[int, list[int]],
    ignore_eos: bool = False,
    include_prompt: bool = False,
) -> list[dict[str, Any]]:
    """Loads and runs the model on cpu with transformers, caching the results.
    Returns cached results if any are found to avoid overhead."""

    if isinstance(model, ModelInfo):
        revision = model.revision
        model_name = model.name
    else:
        revision = None
        model_name = model

    if not isinstance(max_new_tokens, list):
        max_new_tokens = [max_new_tokens] * len(prompts)

    results = []
    for prompt, max_tokens in zip(prompts, max_new_tokens):
        results.append(HF_RESULT_CACHE.get_cached_result(model, prompt, max_tokens))

    if all(results):
        # Everything hit cache
        return results

    assert os.getenv("GITHUB_ACTIONS", "") != "true", (
        "HF results cache miss during Github Actions run. "
        "Please run tests locally with `-m 'cpu'` and check in the changes "
        "to hf_cache.json"
    )

    hf_model = AutoModelForCausalLM.from_pretrained(model_name, revision=revision)
    hf_tokenizer = AutoTokenizer.from_pretrained(model_name, revision=revision)
    if ignore_eos:
        hf_model.generation_config.eos_token_id = None

    for prompt_index, prompt in enumerate(prompts):
        if results[prompt_index]:
            # Already have cached result
            continue

        hf_input_tokens = (
            hf_tokenizer(prompt, return_tensors="pt").input_ids
            if isinstance(prompt[0], str)
            else torch.tensor([prompts[prompt_index]])
        )
        hf_output = hf_model.generate(
            hf_input_tokens,
            do_sample=False,
            max_new_tokens=max_new_tokens[prompt_index],
            return_dict_in_generate=True,
            output_scores=True,
        )

        # decode output tokens after first removing input tokens (prompt)
        hf_generated_text = hf_tokenizer.batch_decode(
            hf_output.sequences[:, len(hf_input_tokens[0]) :]
        )[0]
        hf_transition_scores = hf_model.compute_transition_scores(
            hf_output.sequences, hf_output.scores, normalize_logits=True
        )

        # return HF generated text, tokens, token ids and logprobs
        result = {}
        result["text"] = hf_generated_text
        result["token_ids"] = []
        result["tokens"] = []
        result["logprobs"] = []
        for tok_index, hf_logprob in enumerate(hf_transition_scores[0]):
            hf_token_id = hf_output.sequences[0][tok_index + len(hf_input_tokens[0])]
            result["token_ids"].append(hf_token_id.item())
            result["tokens"].append(hf_tokenizer.decode(hf_token_id))
            result["logprobs"].append(hf_logprob.item())
        result["token_ids"] = tuple(result["token_ids"])
        result["tokens"] = tuple(result["tokens"])
        result["logprobs"] = tuple(result["logprobs"])
        if include_prompt:
            result["prompt"] = prompt

        # Save and cache new result
        results[prompt_index] = result
        HF_RESULT_CACHE.add_to_cache(model, prompt, max_new_tokens[prompt_index], result)

    # Write back to the cache
    HF_RESULT_CACHE.write_cache()
    return results


# compare results
def compare_results(
    model: str | ModelInfo,
    tensor_parallel_size: int,
    backend: str,
    vllm_results: list[dict[str, Any]],
    hf_results: list[dict[str, Any]],
    prompts: list[str] | None = None,
):
    revision = None
    if isinstance(model, ModelInfo):
        revision = model.revision
        model = model.name

    if prompts is None:
        prompts = [""] * len(vllm_results)

    tokenizer = None
    # Decode the prompts if needed
    for idx in range(len(prompts)):
        prompt = prompts[idx]
        if not all(isinstance(t, int) for t in prompt):
            continue
        tokenizer = get_tokenizer(model, revision=revision) if tokenizer is None else tokenizer
        prompts[idx] = tokenizer.decode(prompt)

    print(f"\nmodel:         {model:s}")
    print(f"tp size:       {tensor_parallel_size}")
    print(f"backend:       {backend:s}")
    print(f"\n#prompts:      {len(prompts):d}")
    print(
        f"#HF results:   {len(hf_results):d}"
        f"{'' if len(hf_results) == len(prompts) else '  ERROR':s}"
    )
    print(
        f"#vLLM results: {len(vllm_results):d}"
        f"{'' if len(vllm_results) == len(prompts) else '  ERROR':s}"
    )
    print()

    assert DISABLE_ASSERTS or len(hf_results) == len(vllm_results)
    assert DISABLE_ASSERTS or len(hf_results) == len(prompts)

    for prompt_index, (prompt, hf_result, vllm_result) in enumerate(
        zip(prompts, hf_results, vllm_results)
    ):
        if "text" in vllm_result:
            err_msg = "" if hf_result["text"] == vllm_result["text"] else "  ERROR"
            print(f"\nprompt {prompt_index:3d}:    {repr(prompt):s}")
            print("generated:")
            print(f"        HF:    {repr(hf_result['text']):s}")
            print(f"        vLLM:  {repr(vllm_result['text']):s}{err_msg}")
            print()

        if isinstance(hf_result["token_ids"], list):
            hf_result["token_ids"] = tuple(hf_result["token_ids"])

        if len(hf_result["tokens"]) > 0:
            print(
                "   token id. token               logprob      "
                "   token id. token               logprob"
            )

            logprob_abs_diff_list = []
            logprob_rel_diff_list = []

            for i, (hf_token_id, hf_logprob, vllm_token_id, vllm_logprob) in enumerate(
                zip(
                    hf_result["token_ids"],
                    hf_result["logprobs"],
                    vllm_result["token_ids"],
                    vllm_result["logprobs"],
                )
            ):
                logprob_abs_diff = math.fabs(hf_logprob - vllm_logprob)
                logprob_abs_diff_list.append(logprob_abs_diff)
                logprob_rel_diff = math.fabs(
                    logprob_abs_diff / max(math.fabs(hf_logprob), math.fabs(vllm_logprob))
                )
                logprob_rel_diff_list.append(logprob_rel_diff)

                hf_token = repr(hf_result["tokens"][i]) if "tokens" in vllm_result else "-"
                vllm_token = repr(vllm_result["tokens"][i]) if "tokens" in vllm_result else "-"
                print(
                    f"HF: {hf_token_id:8d} {hf_token:14s} {hf_logprob:14f}  "
                    f"vLLM: {vllm_token_id:8d} {vllm_token:14s} "
                    f"{vllm_logprob:14f}  ",
                    end="",
                )

                if "FP8" in model:
                    # TODO: Improve this. For now our testing model can be
                    # solved with this logic
                    abs_tol = ISCLOSE_ABS_TOL_QUANTIZATION
                else:
                    abs_tol = ISCLOSE_ABS_TOL

                hf_token_prob = math.exp(hf_logprob)
                vllm_token_prob = math.exp(vllm_logprob)

                if hf_token_id != vllm_token_id:  # different tokens
                    if backend == "sendnn" and math.isclose(
                        hf_token_prob,
                        vllm_token_prob,
                        abs_tol=abs_tol,
                    ):
                        # probably still OK
                        print("DIVERGING")
                        break
                    else:
                        print("ERROR")
                        assert DISABLE_ASSERTS or False
                        break
                else:  # identical tokens
                    if math.isclose(
                        hf_token_prob,
                        vllm_token_prob,
                        abs_tol=abs_tol,
                    ):
                        print()
                    else:
                        prob_diff = abs(hf_token_prob - vllm_token_prob)
                        print(f"ERROR (prob_diff = {prob_diff * 100:.2f}%)")
                        assert DISABLE_ASSERTS or False
                        break

            print()
            print(
                "logprob absolute differences: "
                f"average={np.mean(logprob_abs_diff_list):f}  "
                f"maximum={np.max(logprob_abs_diff_list):f}"
            )
            print(
                "logprob relative differences: "
                f"average={np.mean(logprob_rel_diff_list):f}  "
                f"maximum={np.max(logprob_rel_diff_list):f}"
            )

        if hf_result["token_ids"] != vllm_result["token_ids"]:
            print(hf_result["token_ids"])
            print(vllm_result["token_ids"])
        assert (
            DISABLE_ASSERTS
            or backend == "sendnn"
            or hf_result["token_ids"] == vllm_result["token_ids"]
        ), f"Token ids differ: {hf_result['token_ids']} != {vllm_result['token_ids']}"

        print()


def check_output_against_hf(
    model: str | ModelInfo, backend, max_new_tokens, vllm_results, prompts
) -> None:
    hf_outputs = generate_hf_output(
        model=model,
        prompts=prompts,
        max_new_tokens=max_new_tokens,
        ignore_eos=True,
    )
    compare_results(
        model=model,
        tensor_parallel_size=1,
        backend=backend,
        vllm_results=vllm_results,
        hf_results=hf_outputs,
        prompts=prompts,
    )


# Hugging Face
def st_embeddings(model: str | ModelInfo, prompts: list[str]) -> list[dict[str, Any]]:
    if isinstance(model, ModelInfo):
        model = SentenceTransformer(model.name, revision=model.revision)
    else:
        model = SentenceTransformer(model)

    results = []
    for prompt in prompts:
        embeddings = model.encode(prompt)

        # return ST generated embeddings
        result = {}
        result["embeddings"] = embeddings
        results.append(result)

    return results


# compare results
def compare_embedding_results(
    model: str | ModelInfo,
    prompts: list[str],
    warmup_shapes: EmbeddingWarmupShapes,
    tensor_parallel_size: int,
    backend: str,
    vllm_results: list[dict[str, Any]],
    hf_results: list[dict[str, Any]],
):
    print(f"\nmodel:         {model}")
    print(f"warmup shapes: {warmup_shapes}")
    print(f"tp size:       {tensor_parallel_size}")
    print(f"backend:       {backend:s}")
    print(f"\n#prompts:      {len(prompts):d}")
    print(
        f"#HF results:   {len(hf_results):d}"
        f"{'' if len(hf_results) == len(prompts) else '  ERROR':s}"
    )
    print(
        f"#vLLM results: {len(vllm_results):d}"
        f"{'' if len(vllm_results) == len(prompts) else '  ERROR':s}"
    )
    print()

    assert DISABLE_ASSERTS or len(hf_results) == len(vllm_results)
    assert DISABLE_ASSERTS or len(hf_results) == len(prompts)

    for hf_result, vllm_result in zip(hf_results, vllm_results):
        sim = util.pytorch_cos_sim(hf_result["embeddings"], vllm_result["embeddings"])

        assert math.isclose(sim, 1.0, rel_tol=0.05)


# vLLM / Spyre
def spyre_vllm_embeddings(
    model: str | ModelInfo,
    prompts: list[str],
    max_model_len: int,
    tensor_parallel_size: int,
    backend: str,
) -> list[dict[str, Any]]:
    # NB: This doesn't use the same LLM caching as generate_spyre_vllm_output
    # There aren't as many embedding tests so it's not worth the effort atm to
    # cache

    # Clear any cached decoder model
    LLM_CACHE.clear()

    if isinstance(model, ModelInfo):
        revision = model.revision
        model_name = model.name
    else:
        revision = None
        model_name = model

    vllm_model = LLM(
        model=model_name,
        tokenizer=model_name,
        max_model_len=max_model_len,
        tensor_parallel_size=tensor_parallel_size,
        revision=revision,
        tokenizer_revision=revision,
    )

    vllm_outputs = vllm_model.embed(prompts)

    results = []
    for req_output in vllm_outputs:
        result = {}
        result["embeddings"] = req_output.outputs.embedding
        results.append(result)

    return results


def setup_golden_token(
    model: ModelInfo,
    sampling_params: Union[SamplingParams, list[SamplingParams]],
    hf_outputs: list[dict[str, Any]],
) -> list[SamplingParams]:
    abs_tol = ISCLOSE_ABS_TOL_QUANTIZATION if model.is_quantized else ISCLOSE_ABS_TOL

    if isinstance(sampling_params, SamplingParams):
        # golden tokens injection is per request, so we clone SamplingParams
        sampling_params = [sampling_params.clone() for _ in hf_outputs]

    assert len(sampling_params) == len(hf_outputs)
    for idx, (param, hf) in enumerate(zip(sampling_params, hf_outputs)):
        param.extra_args = {
            "golden_token_injector": {
                "expected_token_ids": hf["token_ids"],
                "expected_logprobs": hf["logprobs"],
                "error_threshold": abs_tol,
                "label": f"#{idx}",
            }
        }
    return sampling_params


def validate_vllm_vs_hf_output(
    model: ModelInfo,
    prompts: Union[list[str], list[list[int]]],
    max_model_len: int,
    max_new_tokens: Union[int, list[int]],
    sampling_params: Union[SamplingParams, list[SamplingParams]],
    tensor_parallel_size: int,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    warmup_shapes: DecodeWarmupShapes | None = None,
    max_num_seqs: int | None = None,
    use_cb: bool = False,
    use_golden_token=True,
    max_num_batched_tokens: int | None = None,
    use_pc: bool = False,
) -> None:
    hf_outputs = generate_hf_output(
        model=model,
        prompts=prompts,
        max_new_tokens=max_new_tokens,
        ignore_eos=True,
    )

    if use_golden_token:
        sampling_params = setup_golden_token(model, sampling_params, hf_outputs)

    vllm_results = generate_spyre_vllm_output(
        model=model,
        prompts=prompts,
        max_model_len=max_model_len,
        sampling_params=sampling_params,
        tensor_parallel_size=tensor_parallel_size,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        use_cb=use_cb,
        use_pc=use_pc,
    )

    compare_results(
        model=model,
        tensor_parallel_size=1,
        backend=backend,
        vllm_results=vllm_results,
        hf_results=hf_outputs,
        prompts=prompts,
    )


# vLLM / Spyre
def generate_spyre_vllm_output(
    model: str | ModelInfo,
    prompts: Union[list[str], list[list[int]]],
    max_model_len: int,
    sampling_params: Union[SamplingParams, list[SamplingParams]],
    tensor_parallel_size: int,
    backend: str,
    monkeypatch: pytest.MonkeyPatch,
    warmup_shapes: DecodeWarmupShapes | None = None,
    max_num_seqs: int | None = None,
    use_cb: bool = False,
    max_num_batched_tokens: int | None = None,
    use_pc: bool = False,
) -> list[dict[str, Any]]:
    # Allows to run multiprocess V1 engine without dumping meaningless logs at
    # shutdown engine this context.
    monkeypatch.setenv("VLLM_SPYRE_OVERRIDE_SIGNALS_HANDLER", "1")

    vllm_model = get_cached_llm(
        model=model,
        max_model_len=max_model_len,
        tensor_parallel_size=tensor_parallel_size,
        backend=backend,
        monkeypatch=monkeypatch,
        warmup_shapes=warmup_shapes,
        max_num_seqs=max_num_seqs,
        max_num_batched_tokens=max_num_batched_tokens,
        use_cb=use_cb,
        use_pc=use_pc,
    )

    vllm_outputs = vllm_model.generate(prompts, sampling_params)
    results = []

    for req_output in vllm_outputs:
        result = extract_output(req_output)
        results.append(result)

    return results


def kwargs_for_mode(mode: str, max_num_seqs: int, warmup_shapes: DecodeWarmupShapes) -> dict:
    """Returns kwargs for validate_vllm_vs_hf_output based on mode"""
    return (
        {
            "max_num_seqs": max_num_seqs,
            "use_cb": True,
            "max_num_batched_tokens": 128 if mode in ["cp", "pc"] else None,
            "use_pc": mode == "pc",
            "warmup_shapes": None,
        }
        if mode != "sb"
        else {
            "warmup_shapes": warmup_shapes,
            "use_cb": False,
            "use_pc": False,
            "max_num_batched_tokens": None,
        }
    )


def extract_output(req_output):
    """Extract text, token_ids, tokens, and logprobs from request output."""

    result = {}
    result["text"] = req_output.outputs[0].text

    # TODO: Workaround for V1, if request does not fit in a warmup shape
    # token_ids may be filled with -1.
    token_ids = [t for t in req_output.outputs[0].token_ids if t >= 0]
    result["token_ids"] = tuple(token_ids)
    result["tokens"] = tuple(
        req_output.outputs[0].logprobs[i][t].decoded_token for i, t in enumerate(token_ids)
    )
    result["logprobs"] = tuple(
        req_output.outputs[0].logprobs[i][t].logprob for i, t in enumerate(token_ids)
    )

    return result


def generate_cache_for_test_swap_decode_programs_for_cb(
    model: str | ModelInfo, prompts: list[str], parent_path: str
):
    """
    This function bakes the generation of prompts with long contexts. Which
    currently are used in the test
    `test_spyre_cb::test_swap_decode_programs_for_cb`.
    """

    # Generate
    assert len(prompts) == 4

    p8k = 8 * 1024
    p16k = 16 * 1024
    p32k = 32 * 1024

    import pickle

    hf_outputs = generate_hf_output(
        model=model,
        prompts=prompts[0:2],
        max_new_tokens=p8k,
        ignore_eos=True,
        include_prompt=True,
    )
    with open(Path(parent_path) / "prompts_8k_bs2.pickle", "wb") as f:
        f.write(pickle.dumps(hf_outputs))

    hf_outputs = generate_hf_output(
        model=model,
        prompts=[prompts[2]],
        max_new_tokens=p16k,
        ignore_eos=True,
        include_prompt=True,
    )
    with open(Path(parent_path) / "prompts_16k_bs1.pickle", "wb") as f:
        f.write(pickle.dumps(hf_outputs))

    hf_outputs = generate_hf_output(
        model=model,
        prompts=[prompts[3]],
        max_new_tokens=p32k,
        ignore_eos=True,
        include_prompt=True,
    )
    with open(Path(parent_path) / "prompts_32k_bs1.pickle", "wb") as f:
        f.write(pickle.dumps(hf_outputs))



"""Compare graphs generated by vLLM-Spyre vs AFTU

Run `python -m pytest tests/aftu/test_compare_graphs.py`.
"""

import os
import sys
import tempfile

import pytest
import torch
from graph_compare_utils import (
    collect_graph_files,
    compare_graphs,
    get_model_path,
    run_inference_py_and_get_graphs,
)
from pytest_mock.plugin import MockerFixture
from spyre_util import DecodeWarmupShapes, ModelInfo, patch_environment
from vllm import LLM

from vllm_spyre.model_executor.model_loader.spyre import SpyreCausalLM


# TODO: AFTU uses float32 for mask, while we use float16, which
# can generates different graphs. For now, until AFTU does not change
# this, we can just mock this method to match with AFTU behavior.
# NOTE: we need to set VLLM_ENABLE_V1_MULTIPROCESSING=0 otherwise this
# mock will not propagate to the child process of the model runner
def mock_get_mask_dtype(mocker: MockerFixture):
    mocker.patch.object(SpyreCausalLM, "get_mask_dtype", return_value=torch.float32)


@pytest.mark.spyre
@pytest.mark.cb
def test_compare_graphs_cb(
    model: ModelInfo,
    max_num_seqs: int,
    max_model_len: int,
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
):
    """Test if the graphs generated by vllm-spyre matches with the ones by
    AFTU in continuous batching mode"""

    model_path = get_model_path(model)

    attn_type = "paged_fp8" if model.is_quantized else "paged"

    if model.is_quantized:
        mock_get_mask_dtype(mocker)

    # fmt: off
    inference_py_args = [
        sys.executable, "-m", "aiu_fms_testing_utils.scripts.inference",
        "--architecture", "hf_pretrained",
        "--model_path", model_path,
        "--tokenizer", model_path,
        "--unfuse_weights",
        "--device_type", "aiu",
        "--compile",
        "--cast_bf16_to_fp16",
        "--compile_dynamic",
        "--min_pad_length", "64",
        "--max_new_tokens", "5",
        "--batch_size", str(max_num_seqs),
        "--compile_dynamic_sendnn",
        "--attention_type", attn_type,
    ]
    # fmt: on

    if not model.is_quantized:
        inference_py_args += ["--default_dtype", "fp16"]

    extra_env = {
        "VLLM_DT_MAX_CONTEXT_LEN": str(max_model_len),
        "VLLM_DT_MAX_BATCH_SIZE": str(max_num_seqs),
        "VLLM_DT_MAX_BATCH_TKV_LIMIT": str(1024 * 128),
    }
    aftu_graphs = run_inference_py_and_get_graphs(inference_py_args, extra_env)

    assert len(aftu_graphs) > 0

    # VLLM

    monkeypatch.setenv("DEE_DUMP_GRAPHS", "vllm_cb")
    # Disable cache to produce the graphs
    monkeypatch.setenv("TORCH_SENDNN_CACHE_ENABLE", "0")

    # need for the mocker
    monkeypatch.setenv("VLLM_ENABLE_V1_MULTIPROCESSING", 0)
    patch_environment(use_cb=True, warmup_shapes=None, backend="sendnn", monkeypatch=monkeypatch)

    # NB: We need AFTU and vllm-spyre to have the same number of paged
    # attention blocks for the graph comparison.
    #
    # vllm-spyre adds 1 block to the full batch count for the padding block and
    # rounds up to a multiple of the batch size (see get_total_spyre_blocks in
    # spyre_model_runner.py).
    #
    # Adjusting the context length for vLLM by the block size counters the
    # padding block adjustment for alignment with AFTU
    adjusted_max_model_len = max_model_len - 64
    original_cwd = os.getcwd()
    try:
        # Change to temp dir to set the test environment clean
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)

            # We only need to load the model
            LLM(
                model=model.name,
                revision=model.revision,
                max_model_len=adjusted_max_model_len,
                tensor_parallel_size=1,
                max_num_seqs=max_num_seqs,
            )

            vllm_graphs = collect_graph_files(tmpdir)
    finally:
        # Restore in case of exception
        os.chdir(original_cwd)

    assert compare_graphs(vllm_graphs, aftu_graphs)


@pytest.mark.spyre
@pytest.mark.parametrize("warmup_shapes", [[(64, 4, 4)]])  # (prompt_length/new_tokens/batch_size)
def test_compare_graphs_static_batching(
    model: ModelInfo,
    warmup_shapes: DecodeWarmupShapes,
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
) -> None:
    """Test if the graphs generated by vllm-spyre matches with the ones by
    AFTU in static batching mode"""

    attn_type = "math_fp8" if model.is_quantized else "sdpa"

    if model.is_quantized:
        mock_get_mask_dtype(mocker)

    model_path = get_model_path(model)

    # fmt: off
    inference_py_args = [
        sys.executable, "-m", "aiu_fms_testing_utils.scripts.inference",
        "--architecture", "hf_pretrained",
        "--model_path", model_path,
        "--tokenizer", model_path,
        "--unfuse_weights",
        "--device_type", "aiu",
        "--compile",
        "--cast_bf16_to_fp16",
        "--compile_dynamic",
        "--fixed_prompt_length", str(warmup_shapes[0][0]),
        "--max_new_tokens", str(warmup_shapes[0][1]),
        "--batch_size", str(warmup_shapes[0][2]),
        "--attention_type", attn_type,
    ]
    # fmt: on

    if not model.is_quantized:
        inference_py_args += ["--default_dtype", "fp16"]

    aftu_graphs = run_inference_py_and_get_graphs(inference_py_args)
    assert len(aftu_graphs) > 0
    # VLLM

    monkeypatch.setenv("DEE_DUMP_GRAPHS", "vllm_sb")
    # Disable cache to produce the graphs
    monkeypatch.setenv("TORCH_SENDNN_CACHE_ENABLE", "0")
    # needed for the mocker
    monkeypatch.setenv("VLLM_ENABLE_V1_MULTIPROCESSING", 0)
    patch_environment(
        use_cb=False, warmup_shapes=warmup_shapes, backend="sendnn", monkeypatch=monkeypatch
    )

    original_cwd = os.getcwd()
    try:
        # Change to temp dir to set the test environment clean
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)

            LLM(
                model=model.name,
                revision=model.revision,
                max_model_len=2048,
                tensor_parallel_size=1,
                max_num_seqs=warmup_shapes[0][2],
            )

            vllm_graphs = collect_graph_files(tmpdir)
    finally:
        # Restore in case of exception
        os.chdir(original_cwd)

    assert compare_graphs(vllm_graphs, aftu_graphs)


@pytest.mark.spyre
@pytest.mark.chunked_prefill
def test_compare_graphs_chunked_prefill(
    model: ModelInfo,
    max_num_seqs: int,
    max_model_len: int,
    monkeypatch: pytest.MonkeyPatch,
):
    """Test if the graphs generated by vllm-spyre matches with the ones by
    AFTU with chunked prefill enabled"""

    model_path = get_model_path(model)

    chunk_size = 128

    # fmt: off
    inference_py_args = [
        sys.executable, "-m", "aiu_fms_testing_utils.scripts.inference",
        "--architecture", "hf_pretrained",
        "--model_path", model_path,
        "--tokenizer", model_path,
        "--unfuse_weights",
        "--device_type", "aiu",
        "--compile",
        "--cast_bf16_to_fp16",
        "--compile_dynamic",
        "--min_pad_length", "64",
        "--max_new_tokens", "5",
        "--batch_size", str(max_num_seqs),
        "--compile_dynamic_sendnn",
        "--prefill_chunk_size", str(chunk_size),
    ]
    # fmt: on

    if not model.is_quantized:
        inference_py_args += ["--default_dtype", "fp16"]
        inference_py_args += ["--attention_type", "paged"]
    else:
        inference_py_args += ["--attention_type", "paged_fp8"]

    extra_env = {
        "VLLM_DT_MAX_CONTEXT_LEN": str(max_model_len),
        "VLLM_DT_MAX_BATCH_SIZE": str(max_num_seqs),
        "VLLM_DT_MAX_BATCH_TKV_LIMIT": str(1024 * 128),
        "VLLM_DT_CHUNK_LEN": str(chunk_size),
    }
    aftu_graphs = run_inference_py_and_get_graphs(inference_py_args, extra_env)

    assert len(aftu_graphs) > 0

    # VLLM

    monkeypatch.setenv("DEE_DUMP_GRAPHS", "vllm_cb_cp")
    # Disable cache to produce the graphs
    monkeypatch.setenv("TORCH_SENDNN_CACHE_ENABLE", "0")

    monkeypatch.setenv("VLLM_DT_CHUNK_LEN", str(chunk_size))
    patch_environment(
        use_cb=True,
        warmup_shapes=None,
        backend="sendnn",
        monkeypatch=monkeypatch,
        use_chunked_prefill=True,
    )

    # NB: We need AFTU and vllm-spyre to have the same number of paged
    # attention blocks for the graph comparison.
    #
    # vllm-spyre adds 1 block to the full batch count for the padding block and
    # rounds up to a multiple of the batch size (see get_total_spyre_blocks in
    # spyre_model_runner.py).
    #
    # Adjusting the context length for vLLM by the block size counters the
    # padding block adjustment for alignment with AFTU
    adjusted_max_model_len = max_model_len - 64
    original_cwd = os.getcwd()
    try:
        # Change to temp dir to set the test environment clean
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)

            # We only need to load the model
            LLM(
                model=model.name,
                revision=model.revision,
                max_model_len=adjusted_max_model_len,
                tensor_parallel_size=1,
                max_num_batched_tokens=chunk_size,
                max_num_seqs=max_num_seqs,
            )

            vllm_graphs = collect_graph_files(tmpdir)
    finally:
        # Restore in case of exception
        os.chdir(original_cwd)

    assert compare_graphs(vllm_graphs, aftu_graphs)
